from airflow.sensors.base_sensor_operator import BaseSensorOperator
from airflow.utils.decorators import apply_defaults
from airflow.models import Variable
import urllib.request
import hashlib
import ssl

class HttpChangeSensor(BaseSensorOperator):
    """
    The HttpChangeSensor can be used to detect a change in a URL
    If the content of the URL changes between various runs of the Sensor the result is True
    If the content remains un-changed the result is False

    The content of the page is read and MD5 hashed, the hash is stored in Airflow variables to remain available between
    runs.
    """

    template_fields = ('url', 'start_tag', 'end_tag')

    @apply_defaults
    def __init__(self,
                 url: str,
                 start_tag: str = None,
                 end_tag: str = None,
                 *args, **kwargs):
        """
        Instantiate an instance of HttpChangeSensor
        :param url: The URL to monitor
        :param start_tag: A string in the URL content from which change detection will start.
            Can be left blank to scan entire URL content
        :param end_tag: A string in the URL content at which change detection should stop.
            Can be left blank
        :param args:
        :param kwargs:
        """
        super(HttpChangeSensor, self).__init__(*args, **kwargs)
        self.url = url
        self.start_tag = start_tag.lower().encode('utf-8') if start_tag else None
        self.end_tag = end_tag.lower().encode('utf-8') if end_tag else None
        self.url_hash = '{}_{}_{}'.format(self.dag_id, self.task_id, hashlib.md5(url.lower().encode('utf-8')).hexdigest())

    def poke(self, context):
        """
        Override of the BaseSensor poke command
        :param context: The execution context
        :return: True if the page changed, False if not
        """
        self.log.info("Checking url {url} to see if data has changed")

        check_result = False

        self.log.info("Loading previous page content hash")
        last_hash = Variable.get(self.url_hash, default_var=None)

        self.log.info("Retrieving page content")
        context = ssl._create_unverified_context()

        page_object = urllib.request.urlopen(self.url, context=context)
        page_content = str(page_object.read()).lower().encode('utf-8')

        self.log.info("Loaded {} bytes of page content".format(len(page_content)))

        start_idx = 0
        end_idx = len(page_content)

        if self.start_tag:
            self.log.info("Start tag specified, scanning...")
            start_idx = page_content.find(self.start_tag)
            if start_idx != -1:
                self.log.info("Located start tag at index: {}".format(start_idx))
                start_idx = start_idx + len(self.start_tag)
                self.log.info('Start_idx calculated as {}'.format(start_idx))

        if self.end_tag:
            self.log.info("End tag specified, scanning...")
            end_idx = page_content.find(self.end_tag)
            if end_idx != -1:
                end_idx = end_idx - 1
            else:
                end_idx = len(page_content)

            self.log.info("Located end tag at index: {}".format(end_idx))

        page_section = page_content[start_idx:end_idx+1]
        self.log.info("Calculating MD5 hash for {} bytes of page section".format(len(page_section)))
        section_hash = hashlib.md5(page_section).hexdigest()

        self.log.info("Section MD5 hash calculated as {}".format(section_hash))

        if last_hash:
            if last_hash != section_hash:
                check_result = True
        else:
            self.log.warn("Last hash instance does not exist")

        self.log.info("Saving current section hash as last hash variable")
        Variable.set(key=self.url_hash, value=section_hash)

        return check_result



import slack
from airflow.hooks.base_hook import BaseHook

from awehflow.alerts.base import Alerter


class SlackAlerter(Alerter):
    def __init__(self, channel, slack_conn_id='slack_default',):
        self.channel = channel
        self.slack_conn_id = slack_conn_id

    def alert(self, context):
        client = slack.WebClient(token=BaseHook.get_connection(self.slack_conn_id).password)

        client.chat_postMessage(
            channel=self.channel,
            text=self.__message_title(context),
            blocks=self.__generate_blocks(context)
        )

    def generate_custom_message(self, project_id, dag_id, message_title, message, title, emoji=':warning:'):
        """
        Builds a custom message dictionary to mimic the format of the default Airflow context dictionary.

        Usage:
        slack = SlackAlerter(channel='#airflow')

        slack.alert(slack.generate_custom_message(project_id='some_project_id',
                                                      dag_id='dag_id',
                                                      message_title='some_message_title',
                                                      message='message',
                                                      title='some_title'))

        :param project_id:
        :param dag_id:
        :param message_title:
        :param message:
        :param emoji:
        :param title:
        :return:
        """
        return {
            'name': title,
            'body': {
                'DAG': dag_id,
                'emoji': emoji,
                message_title: message,
                'project': project_id,
            }
        }

    def __message_title(self, context):
        return context.get('name', 'alert').upper()

    def __message_header(self, context):
        if context.get('name', '') == 'failure':
            emoji = ':rotating_light:'
            title = '<{}|{}>'.format(context['body']['error']['log_url'], self.__message_title(context))
        elif context.get('name', '') == 'success':
            emoji = ':white_check_mark:'
            title = self.__message_title(context)
        else:
            emoji = ':question:'
            if context['body'] and context['body']['emoji']:
                emoji = context['body']['emoji']

            title = self.__message_title(context)

        return {
            "type": "section",
            "text": {
                "type": "mrkdwn",
                "text": "{} *{}*".format(emoji, title)
            }
        }

    def __generate_message_body(self, items_dict):
        return {
            "type": "section",
            "fields": [{
                "type": "mrkdwn",
                "text": "*{}*\n{}".format(key, value)
            } for key, value in items_dict.items()]
        }

    def __message_body(self, context):
        if context['name'] == 'failure':
            items = {
                'DAG': context['body']['name'],
                'Task': context['body']['error']['task_id'],
                'Exception': context['body']['error']['message']
            }
            if 'engineers' in context['body']:
                engineer_handles = []
                for engineer in context['body']['engineers']:
                    if 'slack' in engineer:
                        engineer_handles.append(engineer['slack'])
                if engineer_handles:
                    items['Engineers'] = ', '.join(['<@{}>'.format(handle) for handle in engineer_handles])

        elif context['name'] == 'success':
            items = {key: context['body'][key] for key in context['body'].keys() if
                     key not in ['project', 'engineers', 'status']}
        else:
            items = {key: context['body'][key] for key in context['body'].keys() if
                     key not in ['emoji', 'project']}

        return self.__generate_message_body(items)

    def __message_footer(self, context):
        return {
            "type": "context",
            "elements": [
                {
                    "type": "mrkdwn",
                    "text": ":pushpin: *{}*".format(context['body']['project'])
                }
            ]
        }

    def __generate_blocks(self, context):
        blocks = [
            self.__message_header(context),
            {
                "type": "divider"
            },
            self.__message_body(context)
        ]

        if 'project' in context['body']:
            blocks.append(self.__message_footer(context))

        return blocks



import glob
import json
import os

import pendulum
import yaml
from airflow import DAG
from airflow.utils.trigger_rule import TriggerRule
from jinja2.utils import import_string

from awehflow.config import Config
from awehflow.events.alerts import AlertsEventHandler
from awehflow.operators.flow import StartOperator, SuccessOperator, FailureOperator, EventEmittingOperator
from awehflow.utils import uniquify_name, merge_dicts


class ConfigCache:
    def __init__(self, configs_path):
        self.configs_path = configs_path

    @property
    def cache_directory_path(self):
        return os.path.join(self.configs_path, '.cache')

    @property
    def bundle_path(self):
        return os.path.join(self.cache_directory_path, 'configs_bundle.json')

    @property
    def config_files_glob(self):
        return glob.glob(os.path.join(self.configs_path, '**', '*.yml'), recursive=True)

    def refresh_cache(self):
        """Before deploying to Airflow, always refresh the cache, then sync the bundle to the live Airflow instance"""

        if not os.path.exists(self.cache_directory_path):
            os.makedirs(self.cache_directory_path)

        if os.path.isfile(self.bundle_path):
            os.remove(self.bundle_path)

        configs = []
        for file_path in self.config_files_glob:
            with open(file_path, 'r') as f:
                configs.append(yaml.safe_load(f))

        with open(self.bundle_path, 'w') as fp:
            json.dump(configs, fp)

    def read_cache(self, environment='dev', dag_id_prefix=''):
        if not os.path.exists(self.bundle_path):
            raise FileNotFoundError('Configs cache bundle json file not found')

        with open(self.bundle_path) as json_file:
            configs = json.load(json_file)

        return [Config(config_dict=config, environment=environment, dag_id_prefix=dag_id_prefix) for config in configs]


class DagLoader:
    """Parses pipeline configurations, creates Airflow DAGs and exposes them for Airflow to pickup"""

    def __init__(self, project, configs_path, dag_id_prefix='', event_handlers=None, alerters=None, environment='dev'):
        """
        :param project: Name of the project (e.g. name of repo where instance is instanciated)
        :param configs_path: Full path to folder where config files are located
        :param dag_id_prefix: prefix every dag_id with this
        :param event_handlers: List of EventHandler instances
        :param alerters: List of Alerter instances
        :param environment: prod/dev
        """
        self.project = project
        self.configs_path = configs_path
        self.dag_id_prefix = dag_id_prefix
        self.configs_cache = ConfigCache(configs_path=configs_path)
        self.event_handlers = event_handlers or []
        self.alerters = alerters or []
        self.environment = environment

    def load(self, global_symbol_table):
        configs = self.configs_cache.read_cache(environment=self.environment, dag_id_prefix=self.dag_id_prefix)
        dags = []
        for config in configs:
            variable_name = uniquify_name(config.get('name'))
            dag = self.__build_dag(config)
            dags.append((variable_name, dag))
            global_symbol_table[variable_name] = dag

        return dags

    def __get_dag_event_handlers(self, config: Config):
        event_handlers = self.event_handlers.copy()
        event_handlers.append(AlertsEventHandler(alerters=self.alerters, alert_on=config.get('alert_on')))
        return event_handlers

    def __build_dag(self, config: Config) -> DAG:
        event_handlers = self.__get_dag_event_handlers(config)

        default_dag_args = merge_dicts({
            'owner': config.get('owner'),
            'project': self.project,
            'start_date': config.get('start_date'),
            'end_date': config.get('end_date'),
            'email_on_failure': False,
            'email_on_retry': False,
            'retries': 0,
            'on_failure_callback': self.generate_on_failure_callback(config, event_handlers)
        }, config.get('default_dag_args'))

        dag = DAG(dag_id=config.get('dag_id'),
                  description=config.get('description'),
                  schedule_interval=config.get('schedule'),
                  catchup=config.get('catchup'),
                  template_searchpath=self.configs_path,
                  user_defined_macros=self.generate_user_defined_macros(config),
                  default_args=default_dag_args)

        start = StartOperator(
            task_id='start',
            dag=dag,
            project=self.project,
            job_name=config.get('name'),
            engineers=config.get('engineers'),
            event_handlers=event_handlers
        )

        pre_hook_configs = config.get('pre_hooks', [])
        if pre_hook_configs:
            pre_hooks = DependencySensors(
                task_configs=pre_hook_configs,
                dag=dag,
                params=config.get('params')
            )

            for pre_hook in pre_hooks.leaves:
                pre_hook >> start

        tasks = TasksSubDag(
            task_configs=config.get('tasks'),
            dag=dag,
            job_name=config.get('name'),
            params=config.get('params'),
            event_handlers=event_handlers
        )

        dependency_configs = config.get('dependencies', [])
        if dependency_configs:
            dependencies = DependencySensors(
                task_configs=dependency_configs,
                dag=dag,
                params=config.get('params')
            )

            for dependency_check in dependencies.roots:
                start >> dependency_check

            for dependency_check in dependencies.leaves:
                dependency_check >> tasks.roots
        else:
            start >> tasks.roots

        success = SuccessOperator(
            task_id='success',
            dag=dag,
            project=self.project,
            job_name=config.get('name'),
            engineers=config.get('engineers'),
            trigger_rule=TriggerRule.NONE_FAILED,
            event_handlers=event_handlers
        )

        tasks.leaves >> success

        return dag

    def generate_on_failure_callback(self, config, event_handlers):
        def on_failure_callback(context):
            failure = FailureOperator(
                task_id='failure',
                dag=context.get('dag'),
                project=self.project,
                job_name=config.get('name'),
                engineers=config.get('engineers'),
                event_handlers=event_handlers
            )
            return failure.execute(context=context)
        return on_failure_callback

    @staticmethod
    def generate_user_defined_macros(config):
        return {
            'tz_aware': lambda ds: ds.astimezone(pendulum.timezone(config.get('timezone'))).strftime('%Y-%m-%d'),
            'tz_aware_nodash': lambda ds: ds.astimezone(pendulum.timezone(config.get('timezone'))).strftime('%Y%m%d')
        }

    @staticmethod
    def build_task(task_id, dag, operator, params=None, expected_class=None, event_handlers=None, **kwargs):
        operator_class = import_string(operator)
        if expected_class and not issubclass(operator_class, expected_class):
            raise Exception("Unexpected operator class. Expected {}, got {}".format(expected_class.__name__,
                                                                                    operator_class.__name__))
        obj = operator_class(task_id=task_id, dag=dag, params=(params or {}), **kwargs)
        if event_handlers and isinstance(obj, EventEmittingOperator):
            obj.event_handlers = event_handlers
        return obj


class SubDag:
    """Builds and loads configured sub DAG into parent DAG"""

    def __init__(self, task_configs, dag, params, event_handlers=None):
        self.tasks = {}
        self.dag = dag
        self.__build_sub_dag(task_configs, params, event_handlers)

    @property
    def leaves(self):
        """Tasks with no children. Last to execute"""
        return [self.tasks[task_id] for task_id in self.cached_leaf_ids]

    @property
    def roots(self):
        """Tasks with no parents. First to execute"""
        return [self.tasks[task_id] for task_id in self.cached_root_ids]

    def __build_sub_dag(self, task_configs, params, event_handlers):
        """
        Instantiate all tasks first,
        then set up dependencies
        """

        for task_config in task_configs:
            task = DagLoader.build_task(
                task_id=task_config.get('id'),
                dag=self.dag,
                operator=task_config.get('operator'),
                params=params,
                event_handlers=event_handlers,
                **task_config.get('params', {})
            )
            self.tasks[task.task_id] = task

        for task_config in task_configs:
            upstream = task_config.get('upstream', [])
            downstream = task_config.get('downstream', [])

            for task_id in upstream:
                if task_id in self.tasks:
                    self.tasks[task_config.get('id')] << self.tasks[task_id]

            for task_id in downstream:
                if task_id in self.tasks:
                    self.tasks[task_config.get('id')] >> self.tasks[task_id]

        # Cache leaves and roots. This is necessary since any new dependency being set
        # changes the downstream_list or upstream_list of the tasks
        self.cached_leaf_ids = [task.task_id for task in self.tasks.values() if not task.downstream_list]
        self.cached_root_ids = [task.task_id for task in self.tasks.values() if not task.upstream_list]


class TasksSubDag(SubDag):
    """Builds and loads configured tasks into DAG"""

    def __init__(self, task_configs, dag, job_name, params, event_handlers=None):
        enriched_params = merge_dicts({'job_name': job_name}, params)
        super(TasksSubDag, self).__init__(task_configs, dag, enriched_params, event_handlers)


class DependencySensors(SubDag):
    """Builds and loads configured dependency sensors into DAG"""

    def __init__(self, task_configs, dag, params):
        enriched_task_configs = [merge_dicts({'params': {'timeout': 0}}, task_config) for task_config in task_configs]
        super(DependencySensors, self).__init__(enriched_task_configs, dag, params)



import json
from base64 import b64encode

from airflow.contrib.hooks.gcp_pubsub_hook import PubSubHook

from awehflow.events.base import EventHandler


class PublishToGooglePubSubEventHandler(EventHandler):
    def __init__(self, project_id, topic, gcp_conn_id='google_cloud_default'):
        self.project_id = project_id
        self.topic = topic
        self.pubsub_hook = PubSubHook(gcp_conn_id=gcp_conn_id)

    def catch_all(self, event):
        self.__emit_message(event)

    def __emit_message(self, message={}):
        message = {
            'data': self.__encode_message(message)
        }
        self.__emit_batch([message])

    def __emit_batch(self, messages):
        self.pubsub_hook.publish(self.project_id, self.topic, messages)

    @staticmethod
    def __encode_message(message):
        return b64encode(json.dumps(message).encode()).decode()


import copy
import datetime

import pendulum
from dotty_dict import Dotty

from awehflow.utils import merge_dicts


class Config(Dotty):
    DEFAULTS = {
        'name': None,
        'version': 1,
        'description': None,
        'owner': 'airflow',
        'start_date': None,
        'end_date': None,
        'schedule': None,
        'catchup': False,
        'timezone': 'Africa/Johannesburg',
        'engineers': [],
        'alert_on': [],
        'params': {},
        'pre_hooks': [],
        'dependencies': [],
        'tasks': [],
        'default_dag_args': {}
    }

    REQUIRED = [
        'name'
    ]

    def __init__(self, config_dict, environment='dev', dag_id_prefix=''):
        config = merge_dicts(copy.deepcopy(self.DEFAULTS), config_dict)

        for field in self.REQUIRED:
            if not config.get(field):
                raise Exception('{} required'.format(field))

        config['dag_id'] = "{}{}_v{}".format(dag_id_prefix, config['name'], config['version'])
        config['start_date'] = self.__get_datetime_obj(config.get('start_date'), config.get('timezone'))
        config['end_date'] = self.__get_datetime_obj(config.get('end_date'), config.get('timezone'))
        config['params'] = self.__apply_environment_to_params(environment, config.get('params'))

        super(Config, self).__init__(config)

    @staticmethod
    def __apply_environment_to_params(environment, params):
        tmp_params = copy.deepcopy(params.get('default', {}))
        if environment in params:
            tmp_params = merge_dicts(tmp_params, params.get(environment))
        return tmp_params

    @staticmethod
    def __get_datetime_obj(date_string, timezone):
        if not date_string:
            return None
        return datetime.datetime.strptime(date_string, '%Y-%m-%d').replace(tzinfo=pendulum.timezone(timezone))
